#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "SDL/SDL.h"

typedef struct
{
	int wave;
	float fineTune;
	float detune; //fineTuning for right speaker
	int attack;
	int decay; //0 for off
	float pan; //0:left 1:center 2:right
	float volume; //0-1 range

} instrumentConfig;

typedef struct
{
	double freq;
	int duration;
	/*
	 * Proposal - quarter beat : x = subdivisions
	 * ===============
	 *-4 = whole
	 *-2 = half
	 * 1 = quarter
	 * 2 = eighth
	 * 3 = twelth
	 * 4 = sixteenth
	 * 8 = thirty-secondth
	 */
	int currentSample;
	instrumentConfig * instrument;
} note;

typedef struct
{
	note * currentNote;
	note * stream; //dynamic array
	int length;
	int streamIndex;
	long clock;

	int arpType;
	//SFX????
} channel;

#define SAMPLE_RATE 44100
#define TEMPO 145
#define SAMPLES_PER_BEAT (SAMPLE_RATE * 60 / TEMPO)
#define SIZEOF_ARRAY( arr ) sizeof( arr ) / sizeof( arr[0] )
#define CENTS_CALC(c)  pow( 2, c / 1000.0 )

#define PI 3.14159265
#include "filter.h"

double * frequencyTable;
char noteSeq[4] = {35, 47, 47, 42};
char harmSeq[5] = {47, 49, 52, 51, 47};
char harmLen[5] = {1,  2,  2, 1, 1};
char tripSeq[6] = {61, 59, 66, 59, 61, 59};
char bSeq[8] = {23, 23, 15, 16, 20, 15, 16, 18}; //pseudo offbeat
char melSeq[9] = {59, 61, 63, 61, 59, 61, 59, 59, 61};
char melLen[9] = {2,   2,  1,   2, 2,  2,  2, -2, -2};
char melSeq2[9] = {71, 66, 64, 66, 64, 63, 64, 66, 71};
char melLen2[9] = {2,   2,  1,   2, 2,  2,  2, -2, -2};
char melSeq3[11] = {71, 73, 75, 71, 73, 75, 76, 75, 73, 76, 83};
char melLen3[11] = { 4,  4,  2,  4,  4,  4,  4,  2,  2,  2,  2};

channel channels[8];
#define PAT_COUNT 14
note *patterns[PAT_COUNT];
int patternsLength[PAT_COUNT];

signed char patternTable[8][17] = {
/*kick      */	{0,  0,  0,  1,  1,  1,  1,  0,  1,  1,  1,  1,  1,  1,  1,  0,  0},
/*hihats    */	{0,  0,  0,  0,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  0},
/*bass      */	{3, -1, -1, -1,  3, -1, -1, -1,  3, -1, -1, -1,  3, -1, -1, -1, 12},
/*triplets  */	{4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  0},
/*harm1     */	{5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  0},
/*harm2     */	{7, -1, -1,  7, -1, -1,  7, -1, -1,  7, -1, -1,  7, -1, -1,  7,  0},
/*mel       */	{8, -1,  0,  9,  6, -1,  0,  9,  8, -1,  8, -1,  6, -1,  0,  9, 13},
/*snare     */  {0,  0,  0, 11, 10, 10, 10, 11, 10, 10, 10, 10, 10, 10, 10, 11,  0}};

enum waveform { SINE 	= 0,
		SQUARE 	= 1,
		SAW	= 2,
		TRI 	= 3,
		NOISE 	= 10,
		KICK	= 11,
		SILENCE = 9001 };

enum arp { FORWARD = 0,
	   BACKWARD = 1,
	   RANDOM = 2,
	   PINGPONG = 3,
	   RECURSIVE = 4, //WTF is this
	   STOP = 5 };

enum stereo { LEFT = 1, RIGHT = 2 };
int finishSong=1;
instrumentConfig bassI, triSoftI, pianoI, shortI, muffI, kickI, hhoI, hhcI, snareI, sI;
//define instruments settings here:

void initInstruments()
{
	bassI = (instrumentConfig){SINE, CENTS_CALC(7), CENTS_CALC(2), 180, 100000, 1, 0.32f};
	triSoftI = (instrumentConfig){TRI, 1, 1, 820, 10000, 0.6f, 0.23f};
	pianoI = (instrumentConfig){SQUARE, CENTS_CALC(-3), CENTS_CALC(-4), 290, 65000, 1, 0.08f};
	shortI = (instrumentConfig){SQUARE, 1, 1, 5500, 9000, 1.4f, 0.064f};
	muffI = (instrumentConfig){SQUARE, CENTS_CALC(6), CENTS_CALC(-12), 2500, 1000, 1, 0.05f};
	kickI = (instrumentConfig){KICK,  1, 1,5, 6800, 1, 0.45f};
	hhoI = (instrumentConfig){NOISE, 1, 1, 5, 9500, 1, 0.27f};
	hhcI = (instrumentConfig){NOISE, 1, 1, 5, 970, 1, 0.27f};
	snareI = (instrumentConfig){NOISE, 1, 1, 4, 2300, 1, 0.44f};
	sI = (instrumentConfig){SILENCE, 1, 1, 0, 0, 1, 0};
}

static void initChannel( channel *ch, int size )
{
	if( NULL == (ch->stream = calloc(size, sizeof(note))) ){
		printf("Malloc failure\n");
	}
	ch->length = size;
	ch->clock = 0;
	ch->streamIndex = 0; //change if note is not first
	ch->currentNote = &ch->stream[0];
	ch->arpType = FORWARD;
}

static void initPattern( int pattern, int size ) {

	note * n;
	if( NULL == ( n = calloc(size, sizeof(note))) ){
		printf("malloc failure");
		exit(-1);
	}
	patternsLength[pattern] = size;
	patterns[pattern] = n;;
}

void printNote( note *n)
{
	printf("%g, %i, %i, %p\t", n->freq, n->duration, n->currentSample, n->instrument);
}

note * createNote( double f, int dur, instrumentConfig in)
{
	note * n;
	if( NULL==( n=malloc(sizeof(note)) ))
		printf("the fuck with malloc");

	n->freq = f;
	n->duration = dur;
	n->currentSample = 0;
	n->instrument = & in;
	return n;
}

static void initNoteArray()
{
	int a;

	//pattenr 7
	initPattern(7, 16);
	for( a=0 ; a<16 ; a++)
		patterns[7][a] = (note){ frequencyTable[noteSeq[a]], (a+1) % 2 + 1, 0, &triSoftI};
	channels[0].arpType = RANDOM;

	//pattern 0
	initPattern(0, 1);
	note * silenceBar = createNote(0, -4, sI);
	patterns[0] = silenceBar;

	//pattern 1
	initPattern(1, 4);
	int i=0;
	for(i=0; i<4; i++)
		patterns[1][i] = (note){280, 1, 0, &kickI};

	//pattern 2
	initPattern(2, 12);
	note hho = {1, 2, 0, &hhoI};
	note hhc = {1, 4, 0, &hhcI};
	for( i=0 ; i<4; i++)
	{
		hhc.duration = 3;
		patterns[2][i*3] = hhc;
		hhc.duration = 6;
		patterns[2][i*3+1] = hhc;
		patterns[2][i*3+2] = hho;
	}

	//pattern 10
	initPattern(10, 4);
	note qSilence = {0, 1, 0, &sI};
	note sna = {6000, 1, 0, &snareI};
	for( a=0; a<4; a++ )
		patterns[10][a] = a % 2 == 0 ? qSilence : sna;

	//pattern 11
	initPattern(11, 5);
	note hSilence = {0, -2, 0, &sI};
	note eSilence = {0, 2, 0, &sI}; //eigth
	sna.duration = 2;
	patterns[11][0] = eSilence;
	patterns[11][1] = hSilence;
	for( a=2; a<5; a++ )
		patterns[11][a] = sna;

	//pattern 5
	initPattern(5, SIZEOF_ARRAY( harmSeq ) );
	for( a=0; a<SIZEOF_ARRAY( harmSeq ); a++ )
	{
		note m = { frequencyTable[ harmSeq[a] ], harmLen[a], 0, &shortI};
		patterns[5][a] = m;
	}

	//pattern 3
	int p = SIZEOF_ARRAY(bSeq);
	initPattern(3, p);
	for( a=0; a<p; a++ )
	{
		note m = { frequencyTable[ bSeq[ a ]], -2, 0, &bassI};
		patterns[3][a] = m;
	}

	//pattern 12
	initPattern(12, 1);
	patterns[12][0] = (note){frequencyTable[23], -4, 0, &bassI};

	//pattern 4
	initPattern(4, 2 * SIZEOF_ARRAY(tripSeq) );
	for( a=0; a< 2 * SIZEOF_ARRAY( tripSeq ); a++ )
	{
		note m = { frequencyTable[ tripSeq[a] ], 3, 0, &muffI};
		patterns[4][a] = m;
	}

	//pattern 6
	initPattern(6, SIZEOF_ARRAY(melSeq) );
	for( a=0; a<SIZEOF_ARRAY( melSeq ); a++ )
		patterns[6][a] = (note){ frequencyTable[ melSeq[a] ], melLen[a], 0, &pianoI};

	//pattern 8
	initPattern(8, SIZEOF_ARRAY(melSeq2) );
	for( a=0; a<SIZEOF_ARRAY( melSeq2 ); a++ )
		patterns[8][a] = (note){ frequencyTable[ melSeq2[a] ], melLen2[a], 0, &pianoI};

	//pattern 9
	initPattern(9, SIZEOF_ARRAY(melSeq3) );
	for( a=0; a<SIZEOF_ARRAY( melSeq3 ); a++ )
		patterns[9][a] = (note){ frequencyTable[ melSeq3[a] ], melLen3[a], 0, &pianoI};

	//pattern 13
	initPattern(13, 1);
	patterns[13][0] = (note){frequencyTable[71], -4, 0, &pianoI};

}

static void linkPatterns()
{
	int channelLength[7] = {};
	int r, c;
	for( r=0; r<SIZEOF_ARRAY( patternTable ); r++)
	{
		for( c=0; c<SIZEOF_ARRAY( patternTable[0] ); c++)
		{
			int num = patternTable[r][c];
			if( num == 0)
				channelLength[r]++;
			else if( num != -1 )
			{
	 			if( num > SIZEOF_ARRAY(patterns) )
	 				{ printf("WTF is wrong with your table"); exit(-1); }

				channelLength[r] += patternsLength[num];
	 		}
		}
	}

	int b;
	for( r=0; r<SIZEOF_ARRAY( patternTable ) ; r++ )
	{
		initChannel( &channels[r], channelLength[r] );
		for( c=0 ; c<SIZEOF_ARRAY( patternTable[0] ) ; c++)
		{
			//NOTE - here the stream index functions as the index AFTER last loaded note.
			int num = patternTable[r][c];
			if( num != -1)
			{
				int u;
				int patLen = patternsLength[ num ];
				for( u=0 ; u < patLen ; u++ )
				{
					channels[r].stream[u + channels[r].streamIndex ] = patterns[num][u];
					//printf("%i, %i NUM:%i NOTE: ", r, c, num);
					//printNote( & patterns[num][u] );
					//printf("\n");
				}
				channels[r].streamIndex += patLen;
			}
		}
	}
	int j=0;
	for(;j<SIZEOF_ARRAY(channels);j++)
		channels[j].streamIndex = 0;

	//channel settings
	//channels[5].arpType = RANDOM;
}

static void initFreqTable()
{
	if( NULL == ( frequencyTable = calloc( 89, sizeof(double) ))) {
		printf("Fail init freqtable");
	}

	int a;
	frequencyTable[0] = 0;
	for( a=1; a<89 ; a++ )
	{
		frequencyTable[a] = pow( 2, (a-49)/ (double)12 ) * 440;
		//printf( "%i %e \n", a, frequencyTable[a] );
	}
}

static float noteGenerator( note *nt, int lr )
{
	float wave;
	double period;
	period = SAMPLE_RATE / (nt->freq * nt->instrument->fineTune * (lr==1 ? 1 : nt->instrument->detune)) ;
	switch( nt->instrument->wave )
	{
		case SQUARE: ;
			double modCurrSample = fmod( nt->currentSample, period);
			if( modCurrSample < period / 2)
				wave = 1.0f;
			else
				if( modCurrSample < period )
					wave = -1.0f;
				else
					printf("square wave error");
			break;
		case SAW:
			wave = (float)(2 / period * fmod(nt->currentSample, period) )-1.0f;
			break;
		case TRI:
			wave = fabs( 4 * fmod( nt->currentSample / period, 1 ) - 2 ) - 1;
			break;
		case NOISE:
			wave = rand() / (double) RAND_MAX;
			break;
		case SINE:
			wave = (float)sin( 2 * PI * nt->currentSample / period );
			break;
		case KICK:
			wave = (float)sin( 2 * PI * (nt->freq-(nt->currentSample < 6538 ? nt->currentSample * 180.0 / 6538 : 180) ) * nt->currentSample / SAMPLE_RATE );
			// drop te 80 Hz in 6538 samples
		break;
		case SILENCE:
		default:
			wave = 0.0f;
			break;
	}

	//volume envelope
	instrumentConfig * in = nt->instrument;
	if( nt->currentSample < in->attack )
		wave *= (float)(nt->currentSample) / in->attack;
	else if( in->decay != 0 )
	{
			wave *= ((float)(in->decay + in->attack - nt->currentSample))/ in->decay;
			if( nt->currentSample > in->decay + in->attack)
				wave = 0;
	}

	if(lr == 0)
		wave *= (in->pan/2);
	else
	{
		wave *= (1 - in->pan/2);
		nt->currentSample++;
		//printf("%f %d\n", wave, nt->currentSample );
	}

	wave *= in->volume;

	return wave;
}

static void callback(void *uData, Uint8 *stream, int len)
{
	len /= 2; //16bit
	Sint16* dump = (Sint16*) stream;

	int j;
	for(j=0 ; j<len ; j += 2 )
	{

		float instrf[2] = {0, 0}; //used for limiter
		int k;
		for( k=0; k<SIZEOF_ARRAY(channels); k++)
		{
			note * currentNote = channels[k].currentNote;
			float instr[2];
			instr[0] = noteGenerator( currentNote, 0);
			instr[1] = noteGenerator( currentNote, 1);

			float beatFraction;
			if( currentNote->duration < 0 )
				beatFraction = -1.0f * currentNote->duration;
			else
				beatFraction = 1.0f / currentNote->duration;
//		printf("beatFrac, %f, duration. %d, sample coun %li CHANNEL %i\n", beatFraction, currentNote->duration, channels[k].clock, k);
			long samplesLeft = beatFraction * SAMPLES_PER_BEAT - currentNote->currentSample;

			//post fader (click removal)
			if( samplesLeft < 150 )
			{
				instr[0] *= samplesLeft / 150.0f;
				instr[1] *= samplesLeft / 150.0f;
			}
			if( samplesLeft < 0 )
			{
				channels[k].clock = 0;
				switch( channels[k].arpType )
				{
					case RANDOM:
						channels[k].streamIndex = rand() % channels[k].length;
						break;
					case BACKWARD:
						if( channels[k].streamIndex > 0)
							channels[k].streamIndex--;// = rand() % channels[k].length;
						else
							channels[k].streamIndex = channels[k].length - 1;
						break;
					case FORWARD:
					default:
						if( channels[k].streamIndex < channels[k].length - 1)
							channels[k].streamIndex++;// = rand() % channels[k].length;
						else
						{
							channels[k].streamIndex = 0;
							printf("\nEND: %i\n", k);
							finishSong = 0;
						}
				}

				channels[k].currentNote = & channels[k].stream[ channels[k].streamIndex ];
				channels[k].currentNote->currentSample = 0;

			}

			channels[k].clock++;
			instrf[0] += instr[0];
			instrf[1] += instr[1];

		}

		int p;
		for(p=0 ; p<2 ; p++)
		{
			//filter test
			//printf( "%e\t", instrf[p] );

			instrf[p] = updateFilter( instrf[p], p );
			instrf[p] *= 1.5;
			//limiter
			if( instrf[p] > 0.9f )
				instrf[p] = 0.9f + (instrf[p] - 0.9f) / 16; //16:1 ratio
			int instrI = ( (instrf[p]) / 2.0f * 65536 );
			dump[j+p] = instrI;
			//printf("%i\t", instrI);
		}
		//printf("\n");
	}
}

int main (void)
{
	if( SDL_Init( SDL_INIT_AUDIO ) == -1 )
	{
		printf("fail init SDL: %s\n", SDL_GetError() );
		exit(1);
	}

	SDL_AudioSpec desirable;
	extern void callback( void *uData, Uint8 *stream, int len);

	desirable.freq = 44100;
	desirable.format = AUDIO_S16;
	desirable.channels = 2;
	desirable.samples = 1024; //latency
	desirable.callback = callback;
	desirable.userdata = 0;

	if( SDL_OpenAudio( &desirable, NULL) < 0) {
		fprintf(stderr, "Couldn't open audio: %s\n", SDL_GetError());
		return(-1);
	}


	//Pre rednering
	initFreqTable();
	initInstruments();
	initNoteArray();
	linkPatterns();
	calculateFilterCoeffs( 4200.0f, 1.0f);

	//ACTUAL MUSIC GEN

	SDL_PauseAudio(0); //start audio
	int r;
	r = 0;
	while( finishSong != 0 )
		SDL_Delay( 100 );

	//cleanup
	free( channels[0].stream );

	SDL_CloseAudio();
	SDL_Quit();
	exit(0);
}


